/* ----------------------------------------------------------------------------
 * Raima Database Manager
 *
 * Copyright (c) 2012 Raima Inc.,  All rights reserved.
 *
 * Use of this software, whether in source code format, or in executable,
 * binary object code form, is governed by the Raima LICENSE which
 * is fully described in the LICENSE.TXT file, included within this
 * distribution of files.
 * ----------------------------------------------------------------------------
 */

#include <stdio.h>

/* rdmsql.h is required in all files using RDMSQL */
#include "rdmsql.h"

/* EXAMPLE - SQL11 for C using RDMSQL */
/* This example shows multi-user access to a database using separate threads.
 * Two writer threads are created.  The first begins inserting values into
 * the table.  The second waits until the table contains more than 100 rows
 * and then begins deleting the oldest.  A reader thread is also created
 * that prints out the current oldest and newest entries every second.  Each
 * thread uses regular locking.
 *
 * The following RDM functions are illustrated:
 *      rsqlAllocConn/rsqlFreeConn
 *      rsqlAllocStmt/rsqlFreeStmt
 *      rsqlOpenDB/rsqlCloseDB
 *      rsqlExecDirect
 *      rsqlFetch
 *      rsqlTransStart
 *      rsqlTransCommit
 *      rsqlTransRollback
 *      rsqlGetErrorMsg
 *
 * They are fully documented in http://docs.raima.com/rdm/.
 * 
 * For simplicity, this example does not check all return codes, but good
 * programming practices would dictate that they are checked after each
 * RDMSQL call.
 */
static void drop_database(
    HCONN       hDbc,
    const char *name)
{
    HSTMT hStmt;
    char  commandText[80];

    (void) rsqlAllocStmt(hDbc, &hStmt);

    /* Attempt to drop the database.  If the database does not already exist
     * this will fail and we simply ignore the failure for this sample. */
    sprintf(commandText, "DROP DATABASE %s", name);
    (void) rsqlExecDirect(hStmt, commandText);
    (void) rsqlExecDirect(hStmt, "COMMIT");

    (void) rsqlFreeStmt(hStmt);
}

static void create_this_database(
    HCONN       hDbc,
    const char *name,
    const char *cmds[],
    uint32_t    cmd_length)
{
    RSQL_ERRCODE rsqlCode;
    HSTMT        hStmt;
    char         commandText[80];
    uint16_t     ii;

    (void) rsqlAllocStmt(hDbc, &hStmt);

    sprintf(commandText, "CREATE DATABASE %s", name);
    rsqlCode = rsqlExecDirect(hStmt, commandText);

    for (ii = 0; ii < cmd_length; ii++)
    {
        rsqlCode = rsqlExecDirect(hStmt, cmds[ii]);
        if (rsqlCode != errSUCCESS)
            break;
    }

    if (rsqlCode != errSUCCESS)
        fprintf(stderr, "Error creating the %s database\n", name);

    (void) rsqlExecDirect(hStmt, "COMMIT");

    /* Close the newly created database */
    (void) rsqlCloseDB(hDbc, name);

    (void) rsqlFreeStmt(hStmt);
}

static void create_databases(
    HCONN hDbc)
{
    const char *cmds[] = {
        "CREATE TABLE Table1 (id INTEGER PRIMARY KEY)"
    };

    /* Drop any existing copies of the databases. */
    drop_database(hDbc, "SQL11_db");

    /* Attempt to create the databases */
    create_this_database(hDbc, "SQL11_db", cmds, 1);
}

static int16_t shutdown = 0;

static void inserter_thread_proc(
    void *ctx)
{
    RSQL_ERRCODE rsqlCode;
    HCONN        hDbc;
    HSTMT        hStmt;
    int32_t      param_id;
    int32_t      ii;

    /* Create a connection for this thread */
    (void) rsqlAllocConn(&hDbc);
    (void) rsqlOpenDB(hDbc, "SQL11_db", "s");

    /* Create a statement to do the insert */
    (void) rsqlAllocStmt(hDbc, &hStmt);

    rsqlPrepare(hStmt, "INSERT INTO Table1 VALUES :id");
    rsqlBindNamedParam(hStmt, "id", tINTEGER, &param_id, NULL, NULL);

    /* Loop until a shutdown is requested */
    for (ii = 0; !shutdown; ii++)
    {
        /* Insert th next value */
        param_id = ii;
        rsqlCode = rsqlExecute(hStmt);
        if (rsqlCode == errSUCCESS)
            (void) rsqlTransCommit(hDbc);
        else
            (void) rsqlTransRollback(hDbc, NULL);
    }

    (void) rsqlFreeStmt(hStmt);
}

static void deleter_thread_proc(
    void *ctx)
{
    RSQL_ERRCODE      rsqlCode;
    HCONN             hDbc;
    HSTMT             hChkStmt;
    HSTMT             hMinStmt;
    HSTMT             hDelStmt;
    int32_t           param_id;
    const RSQL_VALUE *chkResult;
    const RSQL_VALUE *minResult;
    uint16_t          numCols;

    /* Create a connection for this thread */
    (void) rsqlAllocConn(&hDbc);
    (void) rsqlOpenDB(hDbc, "SQL11_db", "s");

    /* Create a command to check the number of rows, one to fine the 'smallest'
     * row, and one to delete */
    (void) rsqlAllocStmt(hDbc, &hChkStmt);
    (void) rsqlAllocStmt(hDbc, &hMinStmt);
    (void) rsqlAllocStmt(hDbc, &hDelStmt);

    rsqlCode = rsqlPrepare(hChkStmt, "SELECT COUNT(id) FROM Table1");
    rsqlCode = rsqlPrepare(hMinStmt, "SELECT MIN(id) FROM Table1");
    rsqlCode = rsqlPrepare(hDelStmt, "DELETE FROM Table1 WHERE id = :minid");
    rsqlCode = rsqlBindNamedParam(hDelStmt, "minid", tINTEGER, &param_id, NULL, NULL);

    /* Loop until a shutdown is requested */
    while (!shutdown)
    {
        /* Query the number of rows in the table */
        rsqlCode = rsqlExecute(hChkStmt);
        rsqlCode = rsqlFetch(hChkStmt, &chkResult, &numCols);
        if (rsqlCode == errSUCCESS)
        {
            int32_t count = chkResult[0].vt.lv;

            /* There should be only one row, so close the reader */
            (void) rsqlCloseStmt(hChkStmt);

            if (count > 100)
            {
                /* If there at least 100 rows, get the smallest and delete it */
                rsqlCode = rsqlExecute(hMinStmt);
                rsqlCode = rsqlFetch(hMinStmt, &minResult, &numCols);
                if (rsqlCode == errSUCCESS)
                {
                    param_id = minResult[0].vt.lv;
                    rsqlCode = rsqlExecute(hDelStmt);
                    if (rsqlCode == errSUCCESS)
                        (void) rsqlTransCommit(hDbc);
                    else
                        (void) rsqlTransRollback(hDbc, NULL);

                    /* There should be only one row, so close the reader */
                    (void) rsqlCloseStmt(hMinStmt);
                }
            }
        }
    }

    (void) rsqlFreeStmt(hChkStmt);
    (void) rsqlFreeStmt(hMinStmt);
    (void) rsqlFreeStmt(hDelStmt);

    (void) rsqlCloseDB(hDbc, "SQL11_db");
    (void) rsqlFreeConn(hDbc);
}

static void reader_thread_proc(
    void *ctx)
{
    RSQL_ERRCODE      rsqlCode;
    HCONN             hDbc;
    HSTMT             hStmt;
    char              commandText[80];
    const RSQL_VALUE *result;
    uint16_t          numCols;

    /* Create a connection for this thread */
    (void) rsqlAllocConn(&hDbc);
    (void) rsqlOpenDB(hDbc, "SQL11_db", "s");

    (void) rsqlAllocStmt(hDbc, &hStmt);

    sprintf(commandText, "SELECT MIN(id), MAX(id) FROM Table1");

    while (!shutdown)
    {
        /* Sleep one second */
        sleep(1);

        rsqlCode = rsqlExecDirect(hStmt, commandText);
        if (rsqlCode == errSUCCESS)
        {
            rsqlCode = rsqlFetch(hStmt, &result, &numCols);
            if (rsqlCode == errSUCCESS)
            {
                printf("Current minimum = %d and current maximum = %d\n",
                        result[0].vt.lv, result[1].vt.lv);

                (void) rsqlCloseStmt(hStmt);
            }
        }
    }

    (void) rsqlFreeStmt(hStmt);

    (void) rsqlCloseDB(hDbc, "SQL11_db");
    (void) rsqlFreeConn(hDbc);
}

/* =======================================================================
*/
int32_t main(void)
{
    HCONN        hDbc;
    psp_thread_t i_id;
    psp_thread_t d_id;
    psp_thread_t r_id;

    /* Create the first connection */
    (void) rsqlAllocConn(&hDbc);

    create_databases(hDbc);

    printf("Starting threads\n");
    printf("   Press <Enter> when you are ready to terminate them\n");

    /* Create and start three threads one to insert, one to delete, and one
       to show the current contents. psp_threadBeginEx() is our platform-
       independent thread function */
    psp_threadBeginEx(inserter_thread_proc, 0, NULL, &i_id, 0, NULL);
    psp_threadBeginEx(deleter_thread_proc, 0, NULL, &d_id, 0, NULL);
    psp_threadBeginEx(reader_thread_proc, 0, NULL, &r_id, 0, NULL);

    /* Wait for user to hit <Enter> */
    getchar();

    /* Indicate to the threads that they should shut down */
    shutdown = 1;

    /* Wait for each thread to finish */
    psp_threadJoin(i_id);
    psp_threadJoin(d_id);
    psp_threadJoin(r_id);

    rsqlFreeConn(hDbc);

    return EXIT_SUCCESS;
}

